/***************************************************************************
                          datamonitor.cpp  -  description
                             -------------------
    begin                : Sun Jun 17 2001
    copyright            : (C) 2001 by Roberto Virga
    email                : rvirga@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <unistd.h>

#include <qfile.h>
#include <qfileinfo.h>
#include <qregexp.h>
#include <qtextstream.h>

#include "datamonitor.h"

DataMonitor::DataMonitor(QString url, QObject *parent, const char *name)
            : QObject(parent, name)
{
	baseURL = url;
	interval = 0;

	watch = new KDirWatch(this);
	watch->stopScan();
	watch->addDir(baseURL, true, false);

	connect(watch, SIGNAL(created(const QString&)), this, SLOT(checkFile(const QString&)));
	connect(watch, SIGNAL(deleted(const QString&)), this, SLOT(checkFile(const QString&)));
	connect(watch, SIGNAL(dirty(const QString&)), this, SLOT(checkFile(const QString&)));
}

DataMonitor::~DataMonitor()
{
}

QString DataMonitor::url() const
{
	return(baseURL);
}

int DataMonitor::getInterval() const
{
	return(interval);
}

void DataMonitor::setInterval(int secs)
{
	if(secs > 0)
	{
		interval = secs;
		watch->startScan();
	} else if(secs <= 0)
	{
		interval = 0;
		watch->stopScan();
	}
}

void DataMonitor::checkFile(int index)
{
	bool modified = false;

	// wait 1/10th of a second to make sure the new data is there
	usleep(100000L);

	data_info info = fileInfo(files[index].fileName);

	if(info.exists != files[index].exists)
	{
		modified = true;
		info.ok = info.exists ? parseFile(index) : false;
	}
	else if(info.exists && (info.timestamp != files[index].timestamp))
	{
		modified = true;
		info.ok = parseFile(index);
	}
	else
		info.ok = files[index].ok;

	files[index] = info;

	if(modified) updateData();
}

void DataMonitor::checkFile(const QString& file)
{
	const QFileInfo fileInfo(file);
	const int i = index(fileInfo.fileName());
	if(i >= 0) checkFile(i);
}

void DataMonitor::checkFiles()
{
	bool modified = false;

	for(int i = 0; i < int(files.count()); i++)
	{
		data_info info = fileInfo(files[i].fileName);

		if(info.exists != files[i].exists)
		{
			modified = true;
			info.ok = info.exists ? parseFile(i) : false;
		}
		else if(info.exists && (info.timestamp != files[i].timestamp))
		{
			modified = true;
			info.ok = parseFile(i);
		}
		else
			info.ok = files[i].ok;

		files[i] = info;
	}

	if(modified) updateData();
}

QStringList DataMonitor::readFile(const QString& fileName, bool& ok, const QString& delimiter)
{
	QFile file(fileName);
	QStringList out;

	ok = false;
	if(file.open(IO_ReadOnly))
	{
		const bool delimited = !delimiter.isEmpty();
		QTextStream text(&file);

		while (!text.atEnd())
		{
			QString line = text.readLine().stripWhiteSpace();

			if(delimited && line == delimiter)
				break;

			out.append(line);
		}

		file.close();
		ok = true;
	}

	return(out);
}

bool DataMonitor::findIntEntry(const QString& key, QStringList& list, const QString& sep, int& value)
{
	bool found = false;

	value = 0;
	for(QStringList::Iterator it = list.begin(); it != list.end(); ++it)
	{
		int pos = (*it).find(sep);

		if(pos == -1) continue;

		QString lhs = *it;
		lhs.truncate(pos);
		lhs = lhs.stripWhiteSpace();

		if(key != lhs) continue;

		found = true;

		QString rhs = *it;
		rhs.remove(0, pos + sep.length());

		value = rhs.toInt();

		list.remove(it);

		break;
	}

	return(found);
}

bool DataMonitor::findLongEntry(const QString& key, QStringList& list, const QString& sep, long& value)
{
	bool found = false;

	value = 0L;
	for(QStringList::Iterator it = list.begin(); it != list.end(); ++it)
	{
		int pos = (*it).find(sep);

		if(pos == -1) continue;

		QString lhs = *it;
		lhs.truncate(pos);
		lhs = lhs.stripWhiteSpace();

		if(key != lhs) continue;

		found = true;

		QString rhs = *it;
		rhs.remove(0, pos + sep.length());

		value = rhs.toLong();

		list.remove(it);

		break;
	}

	return(found);
}

bool DataMonitor::findFloatEntry(const QString& key, QStringList& list, const QString& sep, float& value)
{
	bool found = false;

	value = 0.0f;
	for(QStringList::Iterator it = list.begin(); it != list.end(); ++it)
	{
		int pos = (*it).find(sep);

		if(pos == -1) continue;

		QString lhs = *it;
		lhs.truncate(pos);
		lhs = lhs.stripWhiteSpace();

		if(key != lhs) continue;

		found = true;

		QString rhs = *it;
		rhs.remove(0, pos + sep.length());

		value = rhs.toFloat();

		list.remove(it);

		break;
	}

	return(found);
}

bool DataMonitor::findDoubleEntry(const QString& key, QStringList& list, const QString& sep, double& value)
{
	bool found = false;

	value = 0.0;
	for(QStringList::Iterator it = list.begin(); it != list.end(); ++it)
	{
		int pos = (*it).find(sep);

		if(pos == -1) continue;

		QString lhs = *it;
		lhs.truncate(pos);
		lhs = lhs.stripWhiteSpace();

		if(key != lhs) continue;

		found = true;

		QString rhs = *it;
		rhs.remove(0, pos + sep.length());

		value = rhs.toDouble();

		list.remove(it);

		break;
	}

	return(found);
}

bool DataMonitor::findStringEntry(const QString& key, QStringList& list, const QString& sep, QString& value)
{
	bool found = false;

	value = QString::null;
	for(QStringList::Iterator it = list.begin(); it != list.end(); ++it)
	{
		int pos = (*it).find(sep);

		if(pos == -1) continue;

		QString lhs = *it;
		lhs.truncate(pos);
		lhs = lhs.stripWhiteSpace();

		if(key != lhs) continue;

		found = true;

		QString rhs = *it;
		rhs.remove(0, pos + sep.length());

		value = rhs;

		list.remove(it);

		break;
	}

	return(found);
}

void DataMonitor::addFile(const QString& fileName, int& index)
{
	watch->addFile(baseURL + fileName);

	index = files.count();
	files += fileInfo(fileName);

	if(files[index].exists)
		files[index].ok = parseFile(index);
}

int DataMonitor::index(const QString& fileName)
{
	for(int i = 0; i < int(files.count()); i++)
		if(files[i].fileName == fileName) return(i);
	return(-1);
}

QString DataMonitor::fileName(int index)
{
	return((index < 0 || index > int(files.count())) ? QString::null : files[index].fileName);
}

bool DataMonitor::exists(int index)
{
	return((index < 0 || index > int(files.count())) ? false : files[index].exists);
}

bool DataMonitor::exists(const QString& fileName)
{
	return(exists(index(fileName)));
}

QDateTime DataMonitor::timestamp(int index)
{
	return((index < 0 || index > int(files.count())) ? QDateTime() : files[index].timestamp);
}

uint DataMonitor::size(const QString& fileName)
{
	return(size(index(fileName)));
}

uint DataMonitor::size(int index)
{
	return((index < 0 || index > int(files.count())) ? 0 : files[index].size);
}

QDateTime DataMonitor::timestamp(const QString& fileName)
{
	return(timestamp(index(fileName)));
}

bool DataMonitor::isOK(int index)
{
	return((index < 0 || index > int(files.count())) ? false : files[index].ok);
}

bool DataMonitor::isOK(const QString& fileName)
{
	return(isOK(index(fileName)));
}

bool DataMonitor::parseFile(int)
{
	return(false);
}

void DataMonitor::updateData()
{
}

data_info DataMonitor::fileInfo(const QString& fileName)
{
	data_info out;

	out.fileName = fileName;

	QFileInfo fileInfo(baseURL + fileName);
	fileInfo.refresh();

	out.exists = fileInfo.exists();

	if(out.exists) {
		out.timestamp = fileInfo.lastModified();
		out.size = fileInfo.size();
	}

	out.ok = false;

	return(out);
}
