/***************************************************************************
                          ksetispydoc.cpp  -  description
                             -------------------
    begin                : Mon Jun 18 2001
    copyright            : (C) 2001 by Roberto Virga
    email                : rvirga@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qregexp.h>

#include <kapplication.h>
#include <kconfig.h>
#include <kglobal.h>
#include <klocale.h>
#include <krfcdate.h>

#include "ksetispydoc.h"

const QString LinuxSysInfoURL = "/proc/";

const statsURLStruct defaultStatsURLs[N_STATS_URLS] =
{
	{false, "setiathome.ssl.berkeley.edu", "/fcgi-bin/fcgi", "user_stats", ""},
	{false, "setiathome.ssl.berkeley.edu", "/fcgi-bin/fcgi", "user_stats_new", ""},
	{false, "setiathome.ssl.berkeley.edu", "/fcgi-bin/fcgi", "team_lookup", ""},
	{false, "setiathome.ssl.berkeley.edu", "/cgi-bin/cgi", "print_cert", ""}
};

const int N_CERTIFICATES = 7;

const int Certificates[N_CERTIFICATES] =
{ 100, 250, 500, 1000, 2500, 5000, 10000 };

KSetiSpyDoc::KSetiSpyDoc(QObject *parent, const char *name) : QObject(parent, name)
{
	monitors.setAutoDelete(false);
	current = QString::null;

	defaultSysInfoURL = LinuxSysInfoURL;

	SetiClientMonitor::loadCalibrationPresets();

	cacheManager = new CacheManager(this);
	connect(cacheManager, SIGNAL(notify(int, int)), this, SLOT(handleCacheNotify(int, int)));

	stats.email = QString::null;

	statsStartup = statsUpload = false;
	statsInterval = 0;
	statsTimer = new QTimer(this);
	connect(statsTimer, SIGNAL(timeout()), this, SLOT(updateUserStats()));

	for(int i = 0; i < N_STATS_URLS; i++)
		statsURLs[i] = defaultStatsURLs[i];
	
	lastCmd = lastArgs = QString::null;
}

KSetiSpyDoc::~KSetiSpyDoc()
{
	for(QStringList::Iterator name = monitor_names.begin(); name != monitor_names.end(); name++)
	{
		SetiClientMonitor *monitor = monitors.find(*name);
		if(monitor->killOnExit())
			monitor->kill();
	}
}

SetiClientMonitor *KSetiSpyDoc::setiMonitor()
{
	return(current.isEmpty() ? NULL : monitors.find(current));
}

SetiClientMonitor::State KSetiSpyDoc::setiMonitorState()
{
	return(current.isEmpty() ? SetiClientMonitor::No_Data : monitors.find(current)->currentState());
}

QStringList KSetiSpyDoc::profiles()
{
	return(monitor_names);
}

QString KSetiSpyDoc::currentProfile()
{
	return(current);
}

profileInfoStruct KSetiSpyDoc::getProfile(const QString& name)
{
	profileInfoStruct out;

	out.name = name;

	SetiClientMonitor *monitor = name.isEmpty() ? NULL : monitors.find(name);
	if(monitor != NULL)
	{
		out.url = monitor->url();
		out.email = monitor->defaultEMail();
		out.sysInfoURL = monitor->sysInfoURL();
		out.timeout = monitor->clientTimeOut();
		out.interval = monitor->getInterval();

		out.launchOnStartup = monitor->launchOnStartup();
		out.killOnExit = monitor->killOnExit();
		out.keepAlive = monitor->keepAlive();

		out.clientCmd = monitor->command();
		out.clientArgs = monitor->arguments();

		out.usesCache = monitor->usesCache();

		out.usesSETISpy = monitor->usesSETISpyLog();
		out.writesSETISpy = monitor->writesSETISpyLog();
		out.writesCSV = monitor->writesCSVLog();
		out.logURL = monitor->logURL();

		out.bestGaussian = monitor->bestGaussianLog();
		out.returnedGaussians = monitor->returnedGaussiansLog();

		out.calibration = monitor->calibration();
	}
	else
	{
		// not found: set some reasonable defaults
		out.url = QString::null;
		out.email = cacheManager->getEMail();
		out.sysInfoURL = defaultSysInfoURL;
		out.timeout = 2*60;
		out.interval = 30;

		out.launchOnStartup = true;
		out.killOnExit = out.keepAlive = false;
		out.clientCmd = lastCmd;
		out.clientArgs = lastArgs;

		out.usesCache = false;

		out.usesSETISpy = true;
		out.writesSETISpy = out.writesCSV = false;

		const imageLogStruct defaultImageLog =
		{
			imageLogStruct::None,
			0.0,
			imageLogStruct::PNG,
			imageLogStruct::Default,
			QString::null
		};
		out.bestGaussian = out.returnedGaussians = defaultImageLog;

		out.calibration = defaultCalibration();
	}

	return(out);
}

QString KSetiSpyDoc::getProfileName(const SetiClientMonitor *monitor)
{
	for(QStringList::Iterator name = monitor_names.begin(); name != monitor_names.end(); name++)
	{
		SetiClientMonitor *current = monitors.find(*name);
		if(monitor == current)
			return(*name);
	}
	return(QString::null);
}

calibrationDataStruct KSetiSpyDoc::defaultCalibration()
{
	calibrationDataStruct out;

	out.cpu = QString::null;

	const double low_defaults[7] = {5.0, 25.0, 45.0, 50.0, 85.0, 95.0, 100.0};
	out.low.clear();
	for(uint i = 0; i < 7; i++)
	{
		calibrationItemStruct item;

		item.reported = item.actual = low_defaults[i] * 1e-2;
		out.low += item;
	}

	const double medium_defaults[7] = {5.0, 15.0, 45.0, 50.0, 80.0, 90.0, 100.0};
	out.medium.clear();
	for(uint i = 0; i < 7; i++)
	{
		calibrationItemStruct item;

		item.reported = item.actual = medium_defaults[i] * 1e-2;
		out.medium += item;
	}

	const double high_defaults[7] = {1.0, 10.0, 45.0, 50.0, 90.0, 95.0, 100.0};
	out.high.clear();
	for(uint i = 0; i < 7; i++)
	{
		calibrationItemStruct item;

		item.reported = item.actual = high_defaults[i] * 1e-2;
		out.high += item;
	}

	return(out);
}

CacheManager *KSetiSpyDoc::getCacheManager() const
{
	return(cacheManager);
}

bool KSetiSpyDoc::statsAutoStartup() const
{
	return(statsStartup);
}

int KSetiSpyDoc::statsTimerInterval() const
{
	return(statsInterval);
}

bool KSetiSpyDoc::statsAutoUpload() const
{
	return(statsUpload);
}

statsURLStruct KSetiSpyDoc::statsURL(StatsURLType type) const
{
	return statsURLs[type];
}

QString KSetiSpyDoc::statsServer() const
{
	return statsURLs[UserData].host;
}

QString KSetiSpyDoc::userDataURL(const QString& email) const
{
	if(email.isEmpty())
		return QString::null;

	if(statsURLs[UserData].custom)
		return statsURLs[UserData].URL;

	const QString command = QString("http://%1%2?cmd=%3").arg(statsURLs[UserData].host)
	                                                     .arg(statsURLs[UserData].path)
	                                                     .arg(statsURLs[UserData].command);
	const QString args = QString("&email=%1").arg(KURL::encode_string(email));

	return (command + args);
}

QString KSetiSpyDoc::userStatsURL(const QString& email) const
{
	if(email.isEmpty())
		return QString::null;

	if(statsURLs[UserLink].custom)
		return statsURLs[UserLink].URL;

	const QString command = QString("http://%1%2?cmd=%3").arg(statsURLs[UserLink].host)
	                                                     .arg(statsURLs[UserLink].path)
	                                                     .arg(statsURLs[UserLink].command);
	const QString args = QString("&email=%1").arg(KURL::encode_string(email));

	return (command + args);
}

QString KSetiSpyDoc::groupStatsURL(const QString& group) const
{
	if(group.isEmpty())
		return QString::null;

	if(statsURLs[GroupLink].custom)
		return statsURLs[GroupLink].URL;

	const QString command = QString("http://%1%2?cmd=%3").arg(statsURLs[GroupLink].host)
	                                                     .arg(statsURLs[GroupLink].path)
	                                                     .arg(statsURLs[GroupLink].command);
	const QString args = QString("&name=%1").arg(KURL::encode_string(group));

	return (command + args);
}

QString KSetiSpyDoc::certificateURL(const QString& email, int wus) const
{
	if(email.isEmpty())
		return QString::null;

	if(statsURLs[CertificateLink].custom)
		return statsURLs[CertificateLink].URL;

	int i;

	for(i = 0; i < N_CERTIFICATES; i++)
		if(Certificates[i] > wus) break;

	if(i == 0)
		return QString::null;

	const QString command = QString("http://%1%2?cmd=%3").arg(statsURLs[CertificateLink].host)
	                                                     .arg(statsURLs[CertificateLink].path)
	                                                     .arg(statsURLs[CertificateLink].command);
	const QString args = QString("&certnum=%1&email=%2&size=0").arg(Certificates[i-1])
	                                                           .arg(KURL::encode_string(email));
	return(command + args);
}

void KSetiSpyDoc::readConfig()
{
	KConfig *config = kapp->config();

	cacheManager->readConfig(config);

	config->setGroup("KSetiSpy");

	defaultSysInfoURL = config->readEntry("Default system info URL", LinuxSysInfoURL);

	const uint n_profiles = config->readUnsignedNumEntry("Profiles", 0);
	const QString profile = config->readEntry("Current profile");

	for(uint i = 0; i < n_profiles; i++)
	{
		profileInfoStruct profile;

		config->setGroup(QString("Profile %1").arg(i));

		profile.name = config->readEntry("Name");
		profile.url = config->readEntry("URL");
		profile.email = config->readEntry("E-mail");

		profile.sysInfoURL = config->readEntry("System info URL", defaultSysInfoURL);

		profile.timeout = config->readNumEntry("Timeout", 5 * 60);
		profile.interval = config->readNumEntry("Poll interval", 30);

		profile.launchOnStartup = config->readBoolEntry("Launch client on startup", true);
		profile.killOnExit = config->readBoolEntry("Kill client on exit", false);
		profile.keepAlive = config->readBoolEntry("Keep client alive", false);
		profile.clientCmd = config->readEntry("Client executable");
		profile.clientArgs = config->readEntry("Client command line arguments");

		profile.usesCache = config->readBoolEntry("Use cache", false);

		profile.usesSETISpy = config->readBoolEntry("Use SETI Spy log format", true);
		profile.writesSETISpy = config->readBoolEntry("Record WUs in SETI Spy log", false);
		profile.writesCSV = config->readBoolEntry("Record WUs in SetiWatch log", false);
		profile.logURL = config->readEntry("Log file directory URL");

		profile.bestGaussian.filter = imageLogStruct::Filter(config->readNumEntry("Best gaussian log filter", 0));
		profile.bestGaussian.threshold = config->readDoubleNumEntry("Best gaussian log threshold", 2.0);
		profile.bestGaussian.format = imageLogStruct::Format(config->readNumEntry("Best gaussian log image format", 1));
		profile.bestGaussian.size = imageLogStruct::Size(config->readNumEntry("Best gaussian log image size", 0));
		profile.bestGaussian.url = config->readEntry("Best gaussian log URL", profile.logURL);

		profile.returnedGaussians.filter = imageLogStruct::Filter(config->readNumEntry("Returned gaussians log filter", 0));
		profile.returnedGaussians.threshold = config->readDoubleNumEntry("Returned gaussians log threshold", 3.5);
		profile.returnedGaussians.format = imageLogStruct::Format(config->readNumEntry("Returned gaussians log image format", 1));
		profile.returnedGaussians.size = imageLogStruct::Size(config->readNumEntry("Returned gaussians log image size", 0));
		profile.returnedGaussians.url = config->readEntry("Returned gaussians log URL", profile.logURL);

		calibrationItemStruct item;

		QStringList low = QStringList::split(" ", config->readEntry("Calibration (low)"));
		profile.calibration.low.clear();
		for(uint i = 0; i < low.count(); i++)
		{
			if(i % 2 == 0)
				item.reported = low[i].toDouble() * 1e-2;
			else {
				item.actual = low[i].toDouble() * 1e-2;
				profile.calibration.low += item;
			}
		}

		QStringList medium = QStringList::split(" ", config->readEntry("Calibration (medium)"));
		profile.calibration.medium.clear();
		for(uint i = 0; i < medium.count(); i++)
		{
			if(i % 2 == 0)
				item.reported = medium[i].toDouble() * 1e-2;
			else {
				item.actual = medium[i].toDouble() * 1e-2;
				profile.calibration.medium += item;
			}
		}

		QStringList high = QStringList::split(" ", config->readEntry("Calibration (high)"));
		profile.calibration.high.clear();
		for(uint i = 0; i < high.count(); i++)
		{
			if(i % 2 == 0)
				item.reported = high[i].toDouble() * 1e-2;
			else {
				item.actual = high[i].toDouble() * 1e-2;
				profile.calibration.high += item;
			}
		}

		if(!profile.name.isEmpty() && !profile.url.isEmpty())
			addProfile(profile, false);
	}

	setCurrentProfile(profile, false);

	config->setGroup("Last User Stats");

	stats.name = config->readEntry("Name", "");
	stats.email = config->readEntry("E-mail");
	stats.group = config->readEntry("Group");
	stats.n_results = config->readNumEntry("Total results");
	stats.total_cpu = config->readDoubleNumEntry("Total CPU time");
	stats.average_time = config->readDoubleNumEntry("Average time");
	stats.last_result_time = config->readDateTimeEntry("Last result time");
	stats.rank.pos = config->readNumEntry("User rank");
	stats.rank.total = config->readNumEntry("Total users");
	stats.n_rank = config->readNumEntry("Users with same rank");
	stats.country = config->readEntry("Country");

	config->setGroup("Stats Servers");

	for(int i = 0; i < N_STATS_URLS; i++)
	{
		QString name;

		name = QString("Use custom URL %1").arg(i);
		statsURLs[i].custom = config->readBoolEntry(name, defaultStatsURLs[i].custom);

		name = QString("Host %1").arg(i);
		statsURLs[i].host = config->readEntry(name, defaultStatsURLs[i].host);

		name = QString("Path %1").arg(i);
		statsURLs[i].path = config->readEntry(name, defaultStatsURLs[i].path);

		name = QString("Command %1").arg(i);
		statsURLs[i].command = config->readEntry(name, defaultStatsURLs[i].command);

		name = QString("URL %1").arg(i);
		statsURLs[i].URL = config->readEntry(name, defaultStatsURLs[i].URL);
	}

	emit updatedProfiles();
	emit updatedLog();
	emit updated();

	if(!stats.name.isEmpty())
		emit updatedUserStats(&stats);
}

void KSetiSpyDoc::saveConfig()
{
	KConfig *config = kapp->config();

	cacheManager->saveConfig(config);

	config->setGroup("KSetiSpy");

	config->writeEntry("Default system info URL", defaultSysInfoURL);

	config->writeEntry("Profiles", monitor_names.count());
	config->writeEntry("Current profile", current);

	for(uint i = 0; i < monitor_names.count(); i++)
	{
		profileInfoStruct profile = getProfile(monitor_names[i]);

		config->setGroup(QString("Profile %1").arg(i));

		config->writeEntry("Name", profile.name);
		config->writeEntry("URL", profile.url);
		config->writeEntry("E-mail", profile.email);

		config->writeEntry("System info URL", profile.sysInfoURL);

		config->writeEntry("Timeout", profile.timeout);
		config->writeEntry("Poll interval", profile.interval);

		config->writeEntry("Launch client on startup", profile.launchOnStartup);
		config->writeEntry("Kill client on exit", profile.killOnExit);
		config->writeEntry("Keep client alive", profile.keepAlive);
		config->writeEntry("Client executable", profile.clientCmd);
		config->writeEntry("Client command line arguments", profile.clientArgs);

		config->writeEntry("Use cache", profile.usesCache);

		config->writeEntry("Use SETI Spy log format", profile.usesSETISpy);
		config->writeEntry("Record WUs in SETI Spy log", profile.writesSETISpy);
		config->writeEntry("Record WUs in SetiWatch log", profile.writesCSV);
		config->writeEntry("Log file directory URL", profile.logURL);

		config->writeEntry("Best gaussian log filter", profile.bestGaussian.filter);
		config->writeEntry("Best gaussian log threshold", profile.bestGaussian.threshold);
		config->writeEntry("Best gaussian log image format", profile.bestGaussian.format);
		config->writeEntry("Best gaussian log image size", profile.bestGaussian.size);
		config->writeEntry("Best gaussian log URL", profile.bestGaussian.url);

		config->writeEntry("Returned gaussians log filter", profile.returnedGaussians.filter);
		config->writeEntry("Returned gaussians log threshold", profile.returnedGaussians.threshold);
		config->writeEntry("Returned gaussians log image format", profile.returnedGaussians.format);
		config->writeEntry("Returned gaussians log image size", profile.returnedGaussians.size);
		config->writeEntry("Returned gaussians log URL", profile.returnedGaussians.url);

		QString low;
		for(uint i = 0; i < profile.calibration.low.count(); i++)
		{
			low += " " + QString::number(profile.calibration.low[i].reported * 1e2, 'f', 2);
			low += " " + QString::number(profile.calibration.low[i].actual * 1e2, 'f', 2);
		}
		low.stripWhiteSpace();
		config->writeEntry("Calibration (low)", low);

		QString medium;
		for(uint i = 0; i < profile.calibration.medium.count(); i++)
		{
			medium += " " + QString::number(profile.calibration.medium[i].reported * 1e2, 'f', 2);
			medium += " " + QString::number(profile.calibration.medium[i].actual * 1e2, 'f', 2);
		}
		medium.stripWhiteSpace();
		config->writeEntry("Calibration (medium)", medium);

		QString high;
		for(uint i = 0; i < profile.calibration.high.count(); i++)
		{
			high += " " + QString::number(profile.calibration.high[i].reported * 1e2, 'f', 2);
			high += " " + QString::number(profile.calibration.high[i].actual * 1e2, 'f', 2);
		}
		high.stripWhiteSpace();
		config->writeEntry("Calibration (high)", high);
	}

	config->setGroup("Last User Stats");

	config->writeEntry("Name", stats.name);
	config->writeEntry("E-mail", stats.email);
	config->writeEntry("Group", stats.group);
	config->writeEntry("Total results", stats.n_results);
	config->writeEntry("Total CPU time", stats.total_cpu);
	config->writeEntry("Average time", stats.average_time);
	config->writeEntry("Last result time", stats.last_result_time);
	config->writeEntry("User rank", stats.rank.pos);
	config->writeEntry("Total users", stats.rank.total);
	config->writeEntry("Users with same rank", stats.n_rank);
	config->writeEntry("Country", stats.country);

	config->setGroup("Stats Servers");

	for(int i = 0; i < N_STATS_URLS; i++)
	{
		QString name;

		name = QString("Use custom URL %1").arg(i);
		config->writeEntry(name, statsURLs[i].custom);

		name = QString("Host %1").arg(i);
		config->writeEntry(name, statsURLs[i].host);

		name = QString("Path %1").arg(i);
		config->writeEntry(name, statsURLs[i].path);

		name = QString("Command %1").arg(i);
		config->writeEntry(name, statsURLs[i].command);

		name = QString("URL %1").arg(i);
		config->writeEntry(name, statsURLs[i].URL);
	}
}

void KSetiSpyDoc::setCurrentProfile(const QString& name, bool signal)
{
	if(name.isEmpty()) return;

	SetiClientMonitor *monitor = monitors.find(name);
	if(monitor != NULL)
	{
		current = name;
		if(signal)
		{
			emit updatedProfiles();
			emit updatedLog();
			emit updated();
		}
	}
}

void KSetiSpyDoc::setStatsAutoStartup(bool set)
{
	statsStartup = set;
}

void KSetiSpyDoc::setStatsTimerInterval(int hours)
{
	if(hours > 0)
	{
		statsInterval = hours;
		statsTimer->start(hours * 60 * 60 * 1000);
	}
	else
	{
		statsInterval = 0;
		statsTimer->stop();
	}
}

void KSetiSpyDoc::setStatsAutoUpload(bool set)
{
	statsUpload = set;
}

void KSetiSpyDoc::setStatsURL(StatsURLType type, const statsURLStruct& datum)
{
	if(type < 0 || type > N_STATS_URLS) return;

	statsURLs[type] = datum;
}

void KSetiSpyDoc::addProfile(const profileInfoStruct& profile, bool signal)
{
	if(profile.name.isEmpty() || profile.url.isEmpty())
		return;

	monitor_names.append(profile.name);
	monitor_names.sort();

	SetiClientMonitor *monitor = new SetiClientMonitor(profile.url, profile.sysInfoURL, profile.logURL, profile.usesSETISpy, this);

	monitor->setDefaultEMail(profile.usesCache ? cacheManager->getEMail() : profile.email);

	monitor->setClientTimeOut(profile.timeout);
	monitor->setInterval(profile.interval);

	monitor->setLaunchOnStartup(profile.launchOnStartup);
	monitor->setKillOnExit(profile.killOnExit);
	monitor->setKeepAlive(profile.keepAlive);

	monitor->setCommand(profile.clientCmd);
	monitor->setArguments(profile.clientArgs);
	if(!profile.clientCmd.isEmpty())
	{
		lastCmd = profile.clientCmd;
		lastArgs = profile.clientArgs;
	}

	monitor->useCache(profile.usesCache);

	monitor->writeSETISpyLog(profile.writesSETISpy);
	monitor->writeCSVLog(profile.writesCSV);

	monitor->setBestGaussianLog(profile.bestGaussian);
	monitor->setReturnedGaussiansLog(profile.returnedGaussians);

	monitor->setCalibration(profile.calibration);

	monitors.insert(profile.name, monitor);

	connect(monitor, SIGNAL(updated(QObject *)), this, SLOT(handleSetiDataChange(QObject *)));
	connect(monitor, SIGNAL(logUpdated(QObject *)), this, SLOT(handleSetiLogChange(QObject *)));
	connect(monitor, SIGNAL(notify(QObject *, int)), this, SLOT(handleClientNotify(QObject *, int)));

	if(current.isEmpty())
		setCurrentProfile(profile.name, signal);
	else if(signal)
		emit updatedProfiles();

	if(monitor->launchOnStartup())
		monitor->exec();
}

void KSetiSpyDoc::changeProfile(const QString& name, const profileInfoStruct& profile, bool signal)
{
	if(name.isEmpty()) return;

	SetiClientMonitor *monitor = monitors.find(name);

	if(monitor == NULL || profile.name.isEmpty())
		return;

	if(name != profile.name)
	{
		monitor_names.remove(name);
		monitor_names.append(profile.name);

		monitors.remove(name);
		monitors.insert(profile.name, monitor);
		if(name == current) current = profile.name;
	}

	monitor->setDefaultEMail(profile.usesCache ? cacheManager->getEMail() : profile.email);

	monitor->setClientTimeOut(profile.timeout);

	monitor->setLaunchOnStartup(profile.launchOnStartup);
	monitor->setKillOnExit(profile.killOnExit);
	monitor->setKeepAlive(profile.keepAlive);

	monitor->setCommand(profile.clientCmd);
	monitor->setArguments(profile.clientArgs);
	if(!profile.clientCmd.isEmpty())
	{
		lastCmd = profile.clientCmd;
		lastArgs = profile.clientArgs;
	}

	monitor->useCache(profile.usesCache);

	monitor->useSETISpyLog(profile.usesSETISpy);
	monitor->writeSETISpyLog(profile.writesSETISpy);
	monitor->writeCSVLog(profile.writesCSV);
	monitor->setLogURL(profile.logURL);

	monitor->setBestGaussianLog(profile.bestGaussian);
	monitor->setReturnedGaussiansLog(profile.returnedGaussians);

	monitor->setCalibration(profile.calibration);

	if(name != profile.name && signal)
	{
		if(current == profile.name) {
			emit updatedLog();
			emit updated();
		}
		emit updatedProfiles();
	}
}

void KSetiSpyDoc::deleteProfile(const QString& name, bool signal)
{
	if(name.isEmpty()) return;

	bool isCurrent = name == current;
	monitor_names.remove(name);
	if(isCurrent)
		current = (monitor_names.count() > 0) ? monitor_names[0] : QString::null;

	SetiClientMonitor *monitor = monitors.find(name);
	if(monitor != NULL) {
		disconnect(monitor);
		delete monitor;
	}
	monitors.remove(name);

	if(signal)
	{
		if(isCurrent) {
			emit updatedLog();
			emit updated();
		}
		emit updatedProfiles();
	}
}

void KSetiSpyDoc::updateUserStats()
{
	SetiClientMonitor *monitor = setiMonitor();
	const seti_data *data = (monitor != NULL) ? monitor->setiData() : NULL;

	QString email;

	if(data != NULL && !data->user.email.isEmpty())
		email = data->user.email;
	else if (!cacheManager->getEMail().isEmpty())
		email = cacheManager->getEMail();
	else return;

	KURL url(userDataURL(email));

	KIO::TransferJob *job = KIO::get(url, true, false);
	connect(job, SIGNAL(data(KIO::Job *, const QByteArray&)), this, SLOT(slotData(KIO::Job *, const QByteArray&)));
	connect(job, SIGNAL(result(KIO::Job *)), this, SLOT(slotResult(KIO::Job *)));

	cdata = QString::null;
	stats.email = email;
}

void KSetiSpyDoc::handleSetiDataChange(QObject *object)
{
	if(!current.isEmpty() && object == monitors.find(current))
		emit updated();
}

void KSetiSpyDoc::handleSetiLogChange(QObject *object)
{
	if(!current.isEmpty() && object == monitors.find(current))
		emit updatedLog();
}

void KSetiSpyDoc::handleClientNotify(QObject *sender, int notify)
{
	const SetiClientMonitor *monitor = (SetiClientMonitor *) sender;
	const QString profile = getProfileName(monitor);
	QString event, message;

	switch(notify) {
		case SetiClientMonitor::WUCompleted:
			event = "WUCompleted";
			message = i18n("work unit completed");
			break;
		case SetiClientMonitor::InterestingGaussian:
			event = "InterestingGaussian";
			message = i18n("an interesting gaussian has been found");
			break;
		default:
			event = "Generic";
			message = i18n("generic notification #%1.").arg(notify);
	}

	KNotifyClient::event(event, i18n("Profile \"%1\": %2.").arg(message).arg(profile));
}

void KSetiSpyDoc::handleCacheNotify(int notify, int count)
{
	QString event, message;

	switch(notify) {
		case CacheManager::WUsDownloaded:
			event = "WUsDownloaded";
			message = i18n("%1 new work units have been downloaded from the server").arg(count, 0);
			break;
		case CacheManager::ResultsUploaded:
			event = "ResultsUploaded";
			message = i18n("%1 results have been uploaded into the server").arg(count, 0);
			if(statsUpload)
				updateUserStats();
			break;
		default:
			event = "Generic";
			message = i18n("Generic notification #%1.").arg(notify);
	}

	KNotifyClient::event(event, message);
}

void KSetiSpyDoc::slotData(KIO::Job *, const QByteArray &data)
{
	cdata += QString(data);
}

void KSetiSpyDoc::slotResult(KIO::Job *job)
{
	if(job->error()) {
		emit updatedUserStats(NULL);
		return;
	} else
	{
		cdata.replace(QRegExp("&nbsp;"), " ");
		cdata.replace(QRegExp("<br>"), " ");
		QStringList data = QStringList::split(QRegExp("\\s*<[^>]*>\\s*"), cdata);

		int first, last, width, pos;

		pos = data.findIndex("Your credit:");
		if(pos < 0) goto error;

		first = pos+1; // first element, first row

		pos = data.findIndex("Country");
		if(pos < 0) goto error;

		last = pos; // last element, first row

		width = last-first+1;
		if(width <= 0 || last + width > int(data.count())) goto error;

		first = last+1; // first element, second row

		pos = data.findIndex("Last result returned:");
		if(pos < 0 || pos >= int(data.count())-1) goto error;

		stats.last_result_time =  convertLastReturnedTime(data[pos+1]);

		if(pos == first+(width-1)) {
			// one less column than there should be: the nickname is empty
			// set it to a single space so that the link to the page will still appear
			stats.name = " ";
			// shift all other entries back one position
			first--;
		} else
			stats.name = data[first+0];

		stats.n_results = data[first+1].toInt();
		stats.total_cpu = convertTotalTime(data[first+2]);
		stats.average_time = convertAvgTime(data[first+3]);
		stats.country = data[first+(width-1)];

		pos = data.findIndex("You belong to the group named:");
		if(pos < 0 || pos >= int(data.count())-1)
			stats.group = QString::null;
		else
			stats.group = data[pos+1];

		pos = data.findIndex("Your rank out of");
		if(pos < 0 || pos >= int(data.count())-3) goto error;

		stats.rank.total = data[pos+1].toInt();
		stats.rank.pos = data[pos+3].toInt();

		pos = data.findIndex("The number of users who have this rank:");
		if(pos < 0 || pos >= int(data.count())-1) goto error;

		stats.n_rank = data[pos+1].toInt();

		emit updatedUserStats(&stats);
		return;

	error:
		emit updatedUserStats(NULL);
		return;
	}
}

double KSetiSpyDoc::convertTotalTime(const QString& string)
{
	int pos;

	pos = string.find(" years");
	if(pos != -1)
		return(string.left(pos).toDouble() * 365 * 24 * 60 * 60);

	pos = string.find(" days");
	if(pos != -1)
		return(string.left(pos).toDouble() * 24 * 60 * 60);

	pos = string.find(" hr");
	if(pos != -1) {
		double out = string.left(pos).toDouble() * 60 * 60;
		QString tmp = string.mid(pos+3);

		pos = tmp.find(" min");
		if(pos != -1) {
			out += tmp.left(pos).toDouble() * 60;
			tmp = tmp.mid(pos+4);
		}

		pos = tmp.find(" sec");
		if(pos != -1)
			out += tmp.left(pos).toDouble();

		return(out);
	}

	return(0.0);
}

double KSetiSpyDoc::convertAvgTime(const QString& string)
{
	QString tmp = string;
	double out = 0.0;
	int pos;

	pos = tmp.find(" hr");
	if(pos != -1)
	{
		out += tmp.left(pos).toDouble() * 60 * 60;
		tmp = tmp.mid(pos+3);
	}
	pos = tmp.find(" min");
	if(pos != -1)
	{
		out += tmp.left(pos).toDouble() * 60;
		tmp = tmp.mid(pos+4);
	}
	pos = tmp.find(" sec");
	if(pos != -1)
		out += tmp.left(pos).toDouble();

	return(out);
}

QDateTime KSetiSpyDoc::convertLastReturnedTime(const QString& string)
{
	QDateTime out;

	QStringList list = QStringList::split(" ", string);
	if(list.count() != 6) return(out);

	const int year = list[4].toInt();

	const QString monthStr = list[1];
	int month;

	if(monthStr == "Jan") month = 1;
	else if(monthStr == "Feb") month = 2;
	else if(monthStr == "Mar") month = 3;
	else if(monthStr == "Apr") month = 4;
	else if(monthStr == "May") month = 5;
	else if(monthStr == "Jun") month = 6;
	else if(monthStr == "Jul") month = 7;
	else if(monthStr == "Aug") month = 8;
	else if(monthStr == "Sep") month = 9;
	else if(monthStr == "Oct") month = 10;
	else if(monthStr == "Nov") month = 11;
	else /* monthStr == "Dec" */ month = 12;

	int day = list[2].toInt();

	out.setDate(QDate(year, month, day));

	list = QStringList::split(":", list[3]);
	if(list.count() != 3) return(out);

	int hours = list[0].toInt();
	int minutes = list[1].toInt();
	int seconds = list[2].toInt();

	out.setTime(QTime(hours, minutes, seconds));
	out = out.addSecs(KRFCDate::localUTCOffset() * 60);

	return(out);
}
